#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include <device_functions.h>

#include <iostream>
#include <cstdlib>
#include <ctime>
#include <stdio.h>
#include <string>

#include <opencv2\core.hpp>
#include <opencv2\opencv_modules.hpp>
#include <opencv2\opencv.hpp>

using namespace cv;

#define BRUSH_SIZE 5
#define LEVELS 20

__device__ __constant__ uchar3* d_src;

__global__ void oilPaint(uchar3* src, uchar3* dst, const int width, const int height) {
	int CountIntensity[LEVELS];
	int RedAverage[LEVELS];
	int GreenAverage[LEVELS];
	int BlueAverage[LEVELS];

	for (int k = 0; k < LEVELS; k++) {
		CountIntensity[k] = 0;
		RedAverage[k] = 0;
		GreenAverage[k] = 0;
		BlueAverage[k] = 0;
	}

	//2D Index of current thread
	const int y = blockIdx.x * blockDim.x + threadIdx.x;
	const int x = blockIdx.y * blockDim.y + threadIdx.y;

	uchar3 curPix;
	int index;
	int row;
	int i,j;
	int crtX, crtY;

	//check if index is out of bounds 
	if (x < height && y < width) {
		for (i = -BRUSH_SIZE; i <= BRUSH_SIZE; ++i) {//all surrounding above and below pixels 		
			crtX = x + i; // neighbour pixel index
			if (crtX >= 0 && crtX < height) {
				row = crtX  * width;
				for (j = -BRUSH_SIZE; j <= BRUSH_SIZE; ++j) {//all surrounding right/left pixels 		
					crtY = y + j;// neighbour pixel index
					if (crtY >= 0 && crtY < width) { //check if neighbour pixel is out of bounds
						curPix = src[row + crtY];
						index = (((curPix.x + curPix.y + curPix.z) / 3.0) * LEVELS) / 255.0;;

						CountIntensity[index]++;
						BlueAverage[index] += curPix.x;
						GreenAverage[index] += curPix.y;
						RedAverage[index] += curPix.z;
					}
				}
			}
		}

		row = 0;
		index = 0;

		for (i = 0; i < LEVELS; i++) {
			if (CountIntensity[i] > row) {
				row = CountIntensity[i];
				index = i;
			}
		}

		// check for 0
		if (row == 0) { row = 1; }

		curPix.x = BlueAverage[index] / row;;
		curPix.y = GreenAverage[index] / row;
		curPix.z = RedAverage[index] / row;

		dst[x * width + y] = curPix;
	}
}

int main(int argc, char** argv) {
	//check command line arguments 
	if (argc != 2) {
		std::cout << argv[0] << ": invalid number of arguments\n";
		return 1;
	}
	std::string fileName = argv[1];

	//read picture 
	Mat src = imread(fileName);
	if (src.empty()) {
		std::cerr << "File not found/read.";
		std::cin.get();
		return -1;
	}

	//define sizes
	const size_t size = src.rows * src.cols;

	//print picutre 
	imshow("Source image", src);

	unsigned long AAtime = 0, BBtime = 0; //check processing time
	AAtime = getTickCount(); //check processing time

	//device setup 
	/*int deviceNum;
	cudaDeviceProp prop;
	cudaGetDevice(&deviceNum);
	cudaGetDeviceProperties(&prop, deviceNum);
	*/
	int ntpb = floor(sqrt(768));
	//int nblks = (size + ntpb - 1) / ntpb;

	//allocate memory 
	uchar3 *d_dst;
	cudaMalloc<uchar3>(&d_src, size * sizeof(uchar3));
	cudaMalloc<uchar3>(&d_dst, size * sizeof(uchar3));

	//copy to device
	cudaMemcpy(d_src, src.ptr<uchar3>(), size * sizeof(uchar3), cudaMemcpyHostToDevice);

	//kernel call 
	const dim3 block(ntpb, ntpb);
	const dim3 grid(ceil((float)src.cols / block.x), ceil((float)src.rows / block.y));
	oilPaint << <grid, block >> > (d_src, d_dst, src.cols, src.rows);

	cudaDeviceSynchronize();

	Mat dst = Mat::zeros(src.size(), src.type());

	//copy back to host
	cudaMemcpy(dst.ptr(), d_dst, size * sizeof(uchar3), cudaMemcpyDeviceToHost);

	//free memory
	cudaFree(d_dst);
	//cudaFree(d_src);

	BBtime = getTickCount(); //check processing time
	printf("%.2lf sec \n", (BBtime - AAtime) / getTickFrequency()); //check processing time

	imshow("Destination image", dst);

	//save oil painting version 
	std::vector<int> compression_params;
	compression_params.push_back(IMWRITE_PNG_COMPRESSION);
	compression_params.push_back(9);
	imwrite("oilVersion-" + fileName, dst, compression_params);

	waitKey();
	return 0;
}
